#include "emu.h"

static int pv_clocksource;

/* --------------------------------------------------------------------- */

const char *feature_bits[32] = {
    [ KVM_FEATURE_CLOCKSOURCE  ] = "clocksource",
    [ KVM_FEATURE_NOP_IO_DELAY ] = "nop-iodelay",
    [ KVM_FEATURE_MMU_OP       ] = "mmu-op",
};

void pv_init(struct xen_cpu *cpu)
{
    char buf[128];
    struct kvm_cpuid_entry entry;
    uint32_t sig[3];
    uint32_t features;

    entry.function = KVM_CPUID_SIGNATURE;
    real_cpuid(&entry);
    sig[0] = entry.ebx;
    sig[1] = entry.ecx;
    sig[2] = entry.edx;

    entry.function = KVM_CPUID_FEATURES;
    real_cpuid(&entry);
    features = entry.eax;

    snprintf(buf, sizeof(buf), "%s: cpu %d, signature \"%.12s\", features 0x%08x",
	     __FUNCTION__, cpu->id, (char*)sig, features);
    print_bits(1, buf, features, features, feature_bits);

    /* pv clocksource */
    if (features & (1 << KVM_FEATURE_CLOCKSOURCE)) {
	uint64_t wall = EMU_PA(&shared_info.wc_version);
	uint64_t sys  = EMU_PA(&cpu->vcpu_info->time);
	if (0 == cpu->id) {
	    /* register wall clock */
	    printk(1, "%s: register wall clock at 0x%" PRIx64 "\n",
		   __FUNCTION__, wall);
	    if (0 != wrmsrl_safe(MSR_KVM_WALL_CLOCK, wall))
		panic("MSR_KVM_WALL_CLOCK wrmsr failed");
	    printk(1, "%s: v%d %d.%09d\n", __FUNCTION__,
		   shared_info.wc_version,
		   shared_info.wc_sec,
		   shared_info.wc_nsec);
	}
	/* per-cpu system time */
	printk(1, "%s: register vcpu %d clock at 0x%" PRIx64 "\n",
	       __FUNCTION__, cpu->id, sys);
	if (0 != wrmsrl_safe(MSR_KVM_SYSTEM_TIME, sys | 1))
	    panic("MSR_KVM_SYSTEM_TIME wrmsr failed");
	printk(1, "%s: v%d sys %" PRId64 " tsc %" PRId64 " mul %d shift %d\n", __FUNCTION__,
	       cpu->vcpu_info->time.version,
	       cpu->vcpu_info->time.system_time,
	       cpu->vcpu_info->time.tsc_timestamp,
	       cpu->vcpu_info->time.tsc_to_system_mul,
	       cpu->vcpu_info->time.tsc_shift);
	pv_clocksource = 1;
    }

#if 0
    /* This is work-in-progress, don't enable this unless
     * you know what you are doing. */
    if (1 /* features & (1 << KVM_FEATURE_CR3_CACHE) */) {
	uint64_t cache;
	cpu->cr3_cache = get_pages(1, "cr3 cache");
//	cache = EMU_PA(cpu->cr3_cache);    /* physical */
	cache = (uintptr_t)cpu->cr3_cache; /* virtual  */
	printk(1, "%s: register cr3 cache at 0x%" PRIx64 " ...\n",
	       __FUNCTION__, cache);
	if (0 != wrmsrl_safe(KVM_MSR_SET_CR3_CACHE, cache)) {
	    printk(1, "%s: ... FAILED\n", __FUNCTION__);
	} else {
	    printk(1, "%s: ... OK, %d entries\n",
		   __FUNCTION__, cpu->cr3_cache->max_idx);
	}
    }
#endif
}

/* --------------------------------------------------------------------- */

void pv_update_systime(int wakeup)
{
    static int wakeups;
    static int update;
    
    if (pv_clocksource)
	return;

    if (wakeup) {
	/* after halt() -- update clock unconditionally */
	update = 1;
	wakeups++;
    } else {
	/* timer irq -- update only if needed */
	update = (0 == wakeups);
	wakeups = 0;
    }

    /* vmexit to userspace so xenner has a chance to update systime */
    if (update)
	vmexit_nop();
}

/* --------------------------------------------------------------------- */

void pv_write_cr3(struct xen_cpu *cpu, ureg_t cr3)
{
#if 0
    int idx;

    if (cpu->cr3_cache && cpu->cr3_cache->max_idx) {
	for (idx = 0; idx < cpu->cr3_cache->max_idx; idx++) {
	    if (cpu->cr3_cache->entry[idx].guest_cr3 == cr3) {
		/*
		 * Cache-hit: we load the cached host-CR3 value.
		 * This never causes any VM exit. (if it does then the
		 * hypervisor could do nothing with this instruction
		 * and the guest OS would be aborted)
		 */
		vminfo.faults[XEN_FAULT_OTHER_CR3_CACHE_HIT]++;
		write_cr3(cpu->cr3_cache->entry[idx].host_cr3);
		return;
	    }
	}
    }
#endif

    write_cr3(cr3);
    return;
}
